#!/usr/bin/env python3
#
# Copyright (c) 2022 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND VMWARE DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VMWARE BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#
#
##########################################################################################################################################
#
#   Testcases
#
###########################################################################################################################################
###########################################################################################################################################
#
# 1.10.1.7. Verify the BGP Local AS functionality with ECMP on 8 links by adding no-prepend and replace-as command in between eBGP Peers.
#
#################################################################################################################################################

import os
import sys
import time
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

from lib.common_config import (
    start_topology,
    write_test_header,
    create_static_routes,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    step,
    check_address_types,
    check_router_status,
    create_static_routes,
    verify_fib_routes,
)

from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    verify_bgp_rib,
    create_router_bgp,
)
from lib.topojson import build_config_from_json

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]

# Global variables
BGP_CONVERGENCE = False
NETWORK = {"ipv4": "10.1.1.0/32", "ipv6": "10:1::1:0/128"}
NEXT_HOP_IP = {"ipv4": "Null0", "ipv6": "Null0"}


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_local_asn_dot_ecmp.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    global BGP_CONVERGENCE
    global ADDR_TYPES
    ADDR_TYPES = check_address_types()

    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module : Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


##########################################################################################################################################
#
#   Testcases
#
###########################################################################################################################################


def test_verify_bgp_local_as_in_ecmp_EBGP_p0(request):
    """
    Verify the BGP Local AS functionality with ECMP on 8 links by
    adding no-prepend and replace-as command in between eBGP Peers.
    """

    tgen = get_topogen()
    global BGP_CONVERGENCE
    if BGP_CONVERGENCE != True:
        pytest.skip("skipped because of BGP Convergence failure")
    # test case name
    tc_name = request.node.name
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)
    reset_config_on_routers(tgen)

    step("Base config is done as part of JSON")
    dut = "r1"
    for addr_type in ADDR_TYPES:
        # Enable static routes
        input_dict_static_route = {
            "r1": {
                "static_routes": [
                    {"network": NETWORK[addr_type], "next_hop": NEXT_HOP_IP[addr_type]}
                ]
            }
        }

        logger.info("Configure static routes")
        result = create_static_routes(tgen, input_dict_static_route)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("configure redistribute static in Router BGP in R1")
        input_dict_static_route_redist = {
            "r1": {
                "bgp": [
                    {
                        "address_family": {
                            addr_type: {
                                "unicast": {"redistribute": [{"redist_type": "static"}]}
                            }
                        }
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_static_route_redist)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Verify IPv4 and IPv6 static routes received on R1")
        result = verify_rib(tgen, addr_type, "r1", input_dict_static_route)
        assert result is True, "Testcase {}: Failed \n Error: {}".format(
            tc_name, result
        )
        result = verify_bgp_rib(tgen, addr_type, "r1", input_dict_static_route)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )
        result = verify_fib_routes(tgen, addr_type, "r1", input_dict_static_route)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as at R3 towards R2.")
    for addr_type in ADDR_TYPES:
        input_dict_r3_to_r2 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {
                                        "r2": {
                                            "dest_link": {
                                                "r3-link1": {
                                                    "local_asn": {"local_as": "1.110"}
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_r3_to_r2)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as at R3 towards R4.")
    dest_link = {}
    for link_no in range(1, 9):
        link = "r3-link" + str(link_no)
        dest_link[link] = {"local_asn": {"local_as": "1.110"}}
    for addr_type in ADDR_TYPES:
        input_dict_r3_to_r4 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {"r4": {"dest_link": dest_link}}
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_r3_to_r4)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure remote-as at R2 towards R3.")
    for addr_type in ADDR_TYPES:
        input_dict_r2_to_r3 = {
            "r2": {
                "bgp": [
                    {
                        "local_as": "1.200",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {
                                        "r3": {
                                            "dest_link": {
                                                "r2-link1": {
                                                    "local_asn": {"remote_as": "1.110"}
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_r2_to_r3)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure remote-as at R4 towards R3.")
    dest_link = {}
    for link_no in range(1, 9):
        link = "r4-link" + str(link_no)
        dest_link[link] = {"local_asn": {"remote_as": "1.110"}}
    for addr_type in ADDR_TYPES:
        input_dict_r4_to_r3 = {
            "r4": {
                "bgp": [
                    {
                        "local_as": "1.400",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {"r3": {"dest_link": dest_link}}
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_r4_to_r3)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("BGP neighborship is verified by following commands in R3 routers")
    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "BGP convergence :Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    step("Verify IPv4 and IPv6 static routes received on R3 & R4")
    for addr_type in ADDR_TYPES:
        static_routes_input = {
            "r1": {
                "static_routes": [
                    {"network": NETWORK[addr_type], "next_hop": NEXT_HOP_IP[addr_type]}
                ]
            }
        }
        for dut in ["r3", "r4"]:
            result = verify_fib_routes(tgen, addr_type, dut, static_routes_input)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

            result = verify_bgp_rib(tgen, addr_type, dut, static_routes_input)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

    step(
        "Verify that AS-110 is got added in the AS list 1.110 1.200 1.100 by following "
        " commands at R3 router."
    )
    dut = "r3"
    aspath = "1.110 1.200 1.100"
    for addr_type in ADDR_TYPES:
        input_static_r1 = {"r1": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_bgp_rib(tgen, addr_type, dut, input_static_r1, aspath=aspath)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as with no-prepend at R3 towards R2.")
    for addr_type in ADDR_TYPES:
        input_dict_no_prep_r3_to_r2 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {
                                        "r2": {
                                            "dest_link": {
                                                "r3-link1": {
                                                    "local_asn": {
                                                        "local_as": "1.110",
                                                        "no_prepend": True,
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_no_prep_r3_to_r2)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as with no-prepend at R3 towards R4.")
    dest_link = {}
    for link_no in range(1, 9):
        link = "r3-link" + str(link_no)
        dest_link[link] = {"local_asn": {"local_as": "1.110"}}
    for addr_type in ADDR_TYPES:
        input_dict_no_prep_r3_to_r4 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {"r4": {"dest_link": dest_link}}
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_no_prep_r3_to_r4)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("BGP neighborship is verified by following commands in R3 routers")
    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "BGP convergence :Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    dut = "r3"
    aspath = "1.200 1.100"
    for addr_type in ADDR_TYPES:
        input_static_r1 = {"r2": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_bgp_rib(tgen, addr_type, dut, input_static_r1, aspath=aspath)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as with no-prepend and replace-as at R3 towards R2")
    for addr_type in ADDR_TYPES:
        input_dict_no_prep_rep_as_r3_to_r2 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {
                                        "r2": {
                                            "dest_link": {
                                                "r3-link1": {
                                                    "local_asn": {
                                                        "local_as": "1.110",
                                                        "no_prepend": True,
                                                        "replace_as": True,
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_no_prep_rep_as_r3_to_r2)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure local-as with no-prepend and replace-as at R3 towards R4")
    dest_link = {}
    for link_no in range(1, 9):
        link = "r3-link" + str(link_no)
        dest_link[link] = {
            "local_asn": {"local_as": "1.110", "no_prepend": True, "replace_as": True}
        }
    for addr_type in ADDR_TYPES:
        input_dict_no_prep_rep_as_r3_to_r4 = {
            "r3": {
                "bgp": [
                    {
                        "local_as": "1.300",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {"r4": {"dest_link": dest_link}}
                                }
                            }
                        },
                    }
                ]
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_no_prep_rep_as_r3_to_r4)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step("BGP neighborship is verified by following commands in R3 routers")
    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "BGP convergence :Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    dut = "r4"
    aspath = "1.110 1.200 1.100"
    for addr_type in ADDR_TYPES:
        input_static_r1 = {"r1": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_bgp_rib(tgen, addr_type, dut, input_static_r1, aspath=aspath)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Part of NetDEF Topology Tests
#
# Copyright (c) 2023 by 6WIND
#

import os
import sys
import pytest

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import required_linux_kernel_version
from lib.checkping import check_ping

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]


def build_topo(tgen):
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")

    tgen.add_router("c11")
    tgen.add_router("c12")
    tgen.add_router("c21")
    tgen.add_router("c22")
    tgen.add_router("c31")
    tgen.add_router("c32")

    tgen.add_link(tgen.gears["r1"], tgen.gears["r2"], "eth0", "eth0")
    tgen.add_link(tgen.gears["r1"], tgen.gears["r3"], "eth1", "eth0")
    tgen.add_link(tgen.gears["r1"], tgen.gears["c11"], "eth2", "eth0")
    tgen.add_link(tgen.gears["r1"], tgen.gears["c12"], "eth3", "eth0")
    tgen.add_link(tgen.gears["r2"], tgen.gears["c21"], "eth1", "eth0")
    tgen.add_link(tgen.gears["r2"], tgen.gears["c22"], "eth2", "eth0")
    tgen.add_link(tgen.gears["r3"], tgen.gears["c31"], "eth1", "eth0")
    tgen.add_link(tgen.gears["r3"], tgen.gears["c32"], "eth2", "eth0")


def setup_module(mod):
    result = required_linux_kernel_version("5.15")
    if result is not True:
        pytest.skip("Kernel requirements are not met")

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    for rname, router in tgen.routers().items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_STATIC, os.path.join(CWD, "{}/staticd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.gears["r1"].run("sysctl net.vrf.strict_mode=1")
    tgen.gears["r1"].run("ip link add vrf10 type vrf table 10")
    tgen.gears["r1"].run("ip link set vrf10 up")
    tgen.gears["r1"].run("ip link add vrf20 type vrf table 20")
    tgen.gears["r1"].run("ip link set vrf20 up")
    tgen.gears["r1"].run("ip link set eth2 master vrf10")
    tgen.gears["r1"].run("ip link set eth3 master vrf20")

    tgen.gears["r2"].run("sysctl net.vrf.strict_mode=1")
    tgen.gears["r2"].run("ip link add vrf10 type vrf table 10")
    tgen.gears["r2"].run("ip link set vrf10 up")
    tgen.gears["r2"].run("ip link add vrf20 type vrf table 20")
    tgen.gears["r2"].run("ip link set vrf20 up")
    tgen.gears["r2"].run("ip link set eth1 master vrf10")
    tgen.gears["r2"].run("ip link set eth2 master vrf20")

    tgen.gears["r3"].run("sysctl net.vrf.strict_mode=1")
    tgen.gears["r3"].run("ip link add vrf10 type vrf table 10")
    tgen.gears["r3"].run("ip link set vrf10 up")
    tgen.gears["r3"].run("ip link add vrf20 type vrf table 20")
    tgen.gears["r3"].run("ip link set vrf20 up")
    tgen.gears["r3"].run("ip link set eth1 master vrf10")
    tgen.gears["r3"].run("ip link set eth2 master vrf20")

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_ping():
    tgen = get_topogen()

    check_ping("c11", "192.168.2.1", True, 120, 1)
    check_ping("c11", "192.168.3.1", True, 10, 1)
    check_ping("c12", "192.168.2.1", True, 10, 1)
    check_ping("c12", "192.168.3.1", True, 10, 1)
    check_ping("c21", "192.168.3.1", True, 10, 1)
    check_ping("c22", "192.168.3.1", True, 10, 1)


def test_sid_unreachable_nht():
    get_topogen().gears["r1"].vtysh_cmd(
        """
        configure terminal
         no ipv6 route 2001:db8:2:2::/64 2001:db8:12::2
        """
    )
    check_ping("c11", "192.168.2.1", False, 10, 1)


def test_sid_reachable_again_nht():
    get_topogen().gears["r1"].vtysh_cmd(
        """
        configure terminal
         ipv6 route 2001:db8:2:2::/64 2001:db8:12::2
        """
    )
    check_ping("c11", "192.168.2.1", True, 10, 1)


def test_sid_unreachable_bgp_update():
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
        router bgp 65002
         no segment-routing srv6
        exit
        router bgp 65002 vrf vrf10
         address-family ipv4 unicast
          no sid vpn export 1
        """
    )
    check_ping("c11", "192.168.2.1", False, 10, 1)


def test_sid_reachable_again_bgp_update():
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
        router bgp 65002
         segment-routing srv6
          locator default
          exit
        exit
        router bgp 65002 vrf vrf10
         address-family ipv4 unicast
          sid vpn export 1
        """
    )
    check_ping("c11", "192.168.2.1", True, 10, 1)


def test_sid_unreachable_no_router():
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
        no router bgp 65002 vrf vrf10
        """
    )
    check_ping("c11", "192.168.2.1", False, 10, 1)


def test_sid_reachable_again_no_router():
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
        router bgp 65002 vrf vrf10
        bgp router-id 192.0.2.2
        !
         address-family ipv4 unicast
          redistribute connected
          sid vpn export 1
          rd vpn export 65002:10
          rt vpn both 0:10
          import vpn
          export vpn
         exit-address-family
        """
    )
    check_ping("c11", "192.168.2.1", True, 10, 1)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_ospf_multi_vrf_bgp_route_leak.py
#
# Copyright (c) 2022 ATCorp
# Jafar Al-Gharaibeh
#

import os
import sys
from functools import partial
import pytest

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger


"""
test_ospf_multi_vrf_bgp_route_leak.py: Test OSPF with multi vrf setup and route leaking.
"""

TOPOLOGY = """
   bgp route leaking (connected/ospf), vrfs:  neno <==> default <==> ray
   routes leaking to vrfs are limited to neno and ray routes.

                      10.0.1.1/24
                            ^
                            |vrf:default
                        +---+---+
      10.0.30.0/24    .1|       |.1
      +-----------------+  R1   +
      |        vrf:neno |       |
      |                 +---+---+                ^
      |.3                 .1|vrf:default         | 10.0.4.4/24
  +---+---+                 |                +---+---+
  |       |            10.0.20.0/24          |       |
  |  R3   |                 |                |  R4   |
  |       |                 |.2              |       |
  +---+---+             +---+---+            +---+---+
      |                 |       | vrf:ray        |.4
      v                 |  R2   +----------------+
10.0.3.3/24             |       |.2         10.0.40.0/24
                        +---+---+
                            |
                            v
                      10.0.2.2/24

"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.ospfd, pytest.mark.bgpd]


def build_topo(tgen):
    "Build function"

    # Create 4 routers
    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    # Create a empty network for router 1
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])

    # Create a empty network for router 2
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])

    # Create a empty network for router 3
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r3"])

    # Create a empty network for router 4
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["r4"])

    # Interconect router 1, 2
    switch = tgen.add_switch("s1-2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 1, 3
    switch = tgen.add_switch("s1-3")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])

    # Interconect router 2, 4
    switch = tgen.add_switch("s2-4")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    logger.info("OSPF Multi VRF Topology with BGP route leaking:\n {}".format(TOPOLOGY))

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    r1_vrf_setup_cmds = [
        "ip link add name neno type vrf table 11",
        "ip link set dev neno up",
        "ip link set dev r1-eth2 vrf neno up",
    ]
    r2_vrf_setup_cmds = [
        "ip link add name ray type vrf table 11",
        "ip link set dev ray up",
        "ip link set dev r2-eth2 vrf ray up",
    ]

    # Starting Routers
    router_list = tgen.routers()

    # Create VRFs on r1/r2 and bind to interfaces
    for cmd in r1_vrf_setup_cmds:
        tgen.net["r1"].cmd(cmd)
    for cmd in r2_vrf_setup_cmds:
        tgen.net["r2"].cmd(cmd)

    logger.info("Testing OSPF VRF support")

    for rname, router in router_list.items():
        logger.info("Loading router %s" % rname)
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    # Initialize all routers.
    tgen.start_router()
    for router in router_list.values():
        if router.has_version("<", "4.0"):
            tgen.set_error("unsupported version")


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


# Shared test function to validate expected output.
def compare_show_ip_route_vrf(rname, expected, vrf_name):
    """
    Calls 'show ip route vrf [vrf_name] route' and compare the obtained
    result with the expected output.
    """
    tgen = get_topogen()
    current = topotest.ip4_route_zebra(tgen.gears[rname], vrf_name)
    ret = topotest.difflines(
        current, expected, title1="Current output", title2="Expected output"
    )
    return ret


def test_ospf_convergence():
    "Test OSPF daemon convergence"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    for rname, router in tgen.routers().items():
        logger.info('Waiting for router "%s" convergence', rname)

        for vrf in ["default", "neno", "ray"]:
            # Load expected results from the command
            reffile = os.path.join(CWD, "{}/ospf-vrf-{}.txt".format(rname, vrf))
            if vrf == "default" or os.path.exists(reffile):
                expected = open(reffile).read()

                # Run test function until we get an result. Wait at most 80 seconds.
                test_func = partial(
                    topotest.router_output_cmp,
                    router,
                    "show ip ospf vrf {} route".format(vrf),
                    expected,
                )
                result, diff = topotest.run_and_expect(test_func, "", count=80, wait=1)
                assertmsg = "OSPF did not converge on {}:\n{}".format(rname, diff)
                assert result, assertmsg


def test_ospf_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv4 kernel routes in "%s"', router.name)
        for vrf in ["default", "neno", "ray"]:
            reffile = os.path.join(CWD, "{}/zebra-vrf-{}.txt".format(router.name, vrf))
            if vrf == "default" or os.path.exists(reffile):
                expected = open(reffile).read()
                # Run test function until we get an result. Wait at most 80 seconds.
                test_func = partial(
                    compare_show_ip_route_vrf, router.name, expected, vrf
                )
                result, diff = topotest.run_and_expect(test_func, "", count=80, wait=1)
                assertmsg = 'OSPF IPv4 route mismatch in router "{}": {}'.format(
                    router.name, diff
                )
                assert result, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# test_ospf_prefix_p2mp_broadcast.py
#
# Copyright (c) 2024 LabN Consulting
# Acee Lindem
#

import os
import sys
from time import sleep
from functools import partial
import pytest

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger

from lib.common_config import (
    step,
)


"""
test_ospf_p2mp_broadcast.py: Test OSPF Point-to-multipoint
"""

TOPOLOGY = """
            +-----+             +-----+
10.1.1.0/24 | r1  |             | r2  | 10.1.2.0/24
 -----------+     |             |     +----------
            +--+--+             +--+--+
               |    10.1.0.0/24    |
               |     +-------+     |
               +---- |       |-----+
                     | P2MP  |
               +---- |       |-----+
               |     +-------+     |
               |                   |
               |                   |
            +--+--+              +-+---+
10.1.3.0/24 | r3  |              | r4  | 10.1.4.0/24
 -----------+     |              |     +----------
            +-----+              +-----+


"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.ospfd, pytest.mark.bgpd]


def build_topo(tgen):
    "Build function"

    # Create 4 routers
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")
    tgen.add_router("r4")

    # Interconect them all to the P2MP network
    switch = tgen.add_switch("s0-p2mp")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r4"])

    # Add standalone network to router 1
    switch = tgen.add_switch("s-r1-1")
    switch.add_link(tgen.gears["r1"])

    # Add standalone network to router 2
    switch = tgen.add_switch("s-r2-1")
    switch.add_link(tgen.gears["r2"])

    # Add standalone network to router 3
    switch = tgen.add_switch("s-r3-1")
    switch.add_link(tgen.gears["r3"])

    # Add standalone network to router 4
    switch = tgen.add_switch("s-r4-1")
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    logger.info("OSPF Point-to-MultiPoint:\n {}".format(TOPOLOGY))

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    # Starting Routers
    router_list = tgen.routers()

    for rname, router in router_list.items():
        logger.info("Loading router %s" % rname)
        router.load_frr_config(os.path.join(CWD, "{}/frr-p2mp.conf".format(rname)))

    # Initialize all routers.
    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def verify_p2mp_interface(
    tgen, router, nbr_cnt, nbr_adj_cnt, delay_reflood, nbr_filter
):
    "Verify the P2MP Configuration and interface settings"

    topo_router = tgen.gears[router]

    step("Test running configuration for P2MP configuration")
    rc = 0
    rc, _, _ = tgen.net[router].cmd_status(
        "show running ospfd | grep 'ip ospf network point-to-multipoint'", warn=False
    )
    assertmsg = (
        "'ip ospf network point-to-multipoint' applied, but not present in "
        + router
        + "configuration"
    )
    assert rc, assertmsg

    step("Test OSPF interface for P2MP settings")
    input_dict = {
        "interfaces": {
            "r1-eth0": {
                "ospfEnabled": True,
                "ipAddress": "10.1.0.1",
                "ipAddressPrefixlen": 24,
                "ospfIfType": "Broadcast",
                "area": "0.0.0.0",
                "routerId": "1.1.1.1",
                "networkType": "POINTOMULTIPOINT",
                "cost": 10,
                "state": "Point-To-Point",
                "opaqueCapable": True,
                "nbrCount": nbr_cnt,
                "nbrAdjacentCount": nbr_adj_cnt,
                "prefixSuppression": False,
                "p2mpDelayReflood": delay_reflood,
                "nbrFilterPrefixList": nbr_filter,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        topo_router,
        "show ip ospf interface r1-eth0 json",
        input_dict,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "P2MP Interface Mismatch on router r1"
    assert result is None, assertmsg


def verify_non_p2mp_interface(tgen):
    "Verify the removal of P2MP Configuration and interface settings"
    r1 = tgen.gears["r1"]

    step("Test running configuration for removal of P2MP configuration")
    rc = 0
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf network point-to-multipoint'", warn=False
    )
    assertmsg = "'ip ospf network point-to-multipoint' not applied, but present in r1 configuration"
    assert rc, assertmsg

    step("Test OSPF interface for default settings")
    input_dict = {
        "interfaces": {
            "r1-eth0": {
                "ospfEnabled": True,
                "ipAddress": "10.1.0.1",
                "ipAddressPrefixlen": 24,
                "ospfIfType": "Broadcast",
                "area": "0.0.0.0",
                "routerId": "1.1.1.1",
                "networkType": "BROADCAST",
                "cost": 10,
                "opaqueCapable": True,
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp, r1, "show ip ospf interface r1-eth0 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "P2MP Interface Mismatch on router r1"
    assert result is None, assertmsg


def verify_p2mp_neighbor(tgen, router, neighbor, state, intf_addr, interface):
    topo_router = tgen.gears[router]

    step("Verify neighbor " + neighbor + " in " + state + " state")
    input_dict = {
        "default": {
            neighbor: [
                {
                    "nbrState": state,
                    "ifaceAddress": intf_addr,
                    "ifaceName": interface,
                }
            ],
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        topo_router,
        "show ip ospf neighbor " + neighbor + " json",
        input_dict,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "P2MP Neighbor " + neighbor + " not in " + state
    assert result is None, assertmsg


def verify_p2mp_neighbor_missing(tgen, router, neighbor):
    topo_router = tgen.gears[router]

    step("Verify neighbor " + neighbor + " missing")
    input_dict = {"default": {}}
    test_func = partial(
        topotest.router_json_cmp,
        topo_router,
        "show ip ospf neighbor " + neighbor + " json",
        input_dict,
        True,  # Require exact match for missing neighbor
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "P2MP Neighbor " + neighbor + " not missing"
    assert result is None, assertmsg


def verify_p2mp_route(tgen, router, prefix, prefix_len, nexthop, interface):
    topo_router = tgen.gears[router]

    step("Verify router " + router + " p2mp route " + prefix + " installed")
    input_dict = {
        prefix: [
            {
                "prefix": prefix,
                "prefixLen": prefix_len,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": nexthop,
                        "interfaceName": interface,
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp,
        topo_router,
        "show ip route " + prefix + " json",
        input_dict,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = prefix + " not installed on router " + router
    assert result is None, assertmsg


def test_p2mp_broadcast_interface():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Verify router r1 interface r1-eth0 p2mp configuration")
    verify_p2mp_interface(tgen, "r1", 3, 3, False, "N/A")

    step("Verify router r1 p2mp interface r1-eth0 neighbors")
    verify_p2mp_neighbor(
        tgen, "r1", "2.2.2.2", "Full/DROther", "10.1.0.2", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "3.3.3.3", "Full/DROther", "10.1.0.3", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "4.4.4.4", "Full/DROther", "10.1.0.4", "r1-eth0:10.1.0.1"
    )

    step("Verify router r1 p2mp routes installed")
    verify_p2mp_route(tgen, "r1", "10.1.2.0/24", 24, "10.1.0.2", "r1-eth0")
    verify_p2mp_route(tgen, "r1", "10.1.3.0/24", 24, "10.1.0.3", "r1-eth0")
    verify_p2mp_route(tgen, "r1", "10.1.4.0/24", 24, "10.1.0.4", "r1-eth0")

    step("Verify router r1 interface r1-eth0 p2mp configuration removal")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nno ip ospf network point-to-multipoint")
    verify_non_p2mp_interface(tgen)

    step("Verify router r1 interface r1-eth0 p2mp configuration application")
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nip ospf network point-to-multipoint")
    verify_p2mp_interface(tgen, "r1", 3, 3, False, "N/A")

    step("Verify restablishment of r1-eth0 p2mp neighbors")
    verify_p2mp_neighbor(
        tgen, "r1", "2.2.2.2", "Full/DROther", "10.1.0.2", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "3.3.3.3", "Full/DROther", "10.1.0.3", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "4.4.4.4", "Full/DROther", "10.1.0.4", "r1-eth0:10.1.0.1"
    )

    step("Verify router r1 p2mp routes reinstalled")
    verify_p2mp_route(tgen, "r1", "10.1.2.0/24", 24, "10.1.0.2", "r1-eth0")
    verify_p2mp_route(tgen, "r1", "10.1.3.0/24", 24, "10.1.0.3", "r1-eth0")
    verify_p2mp_route(tgen, "r1", "10.1.4.0/24", 24, "10.1.0.4", "r1-eth0")


def p2mp_broadcast_neighbor_filter_common(delay_reflood):
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Verify router r1 interface r1-eth0 p2mp configuration")
    verify_p2mp_interface(tgen, "r1", 3, 3, delay_reflood, "N/A")

    step("Verify router r1 p2mp interface r1-eth0 neighbors")
    verify_p2mp_neighbor(
        tgen, "r1", "2.2.2.2", "Full/DROther", "10.1.0.2", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "3.3.3.3", "Full/DROther", "10.1.0.3", "r1-eth0:10.1.0.1"
    )
    verify_p2mp_neighbor(
        tgen, "r1", "4.4.4.4", "Full/DROther", "10.1.0.4", "r1-eth0:10.1.0.1"
    )

    step("Add OSPF interface neighbor-filter to r1")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nip ospf neighbor-filter nbr-filter")

    step("Verify the R1 configuration of 'ip ospf neighbor-filter nbr-filter'")
    neighbor_filter_cfg = (
        tgen.net["r1"]
        .cmd(
            'vtysh -c "show running ospfd" | grep "^ ip ospf neighbor-filter nbr-filter"'
        )
        .rstrip()
    )
    assertmsg = (
        "'ip ospf neighbor-filter nbr-filter' applied, but not present in configuration"
    )
    assert neighbor_filter_cfg == " ip ospf neighbor-filter nbr-filter", assertmsg

    step("Verify non-existent neighbor-filter is not applied to r1 interfaces")
    verify_p2mp_interface(tgen, "r1", 3, 3, delay_reflood, "N/A")

    step("Add nbr-filter prefix-list configuration to r1")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 200 permit any")

    step(
        "Verify neighbor-filter is now applied to r1 interface and neighbors still adjacent"
    )
    verify_p2mp_interface(tgen, "r1", 3, 3, delay_reflood, "nbr-filter")

    step("Add nbr-filter prefix-list configuration to block r4")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 10 deny 10.1.0.4/32")

    step(
        "Verify neighbor-filter is now applied to r1 interface and r4 is no longer adjacent"
    )
    verify_p2mp_interface(tgen, "r1", 2, 2, delay_reflood, "nbr-filter")
    verify_p2mp_neighbor_missing(tgen, "r1", "4.4.4.4")

    step("Verify route to r4 subnet is now through r2")
    verify_p2mp_route(tgen, "r1", "10.1.4.0/24", 24, "10.1.0.2", "r1-eth0")

    step("Add nbr-filter prefix-list configuration to block r2")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 20 deny 10.1.0.2/32")

    step(
        "Verify neighbor-filter is now applied to r1 interface and r2 is no longer adjacent"
    )
    verify_p2mp_interface(tgen, "r1", 1, 1, delay_reflood, "nbr-filter")
    verify_p2mp_neighbor_missing(tgen, "r1", "2.2.2.2")

    step("Verify route to r4 and r2 subnet are now through r3")
    verify_p2mp_route(tgen, "r1", "10.1.2.0/24", 24, "10.1.0.3", "r1-eth0")
    verify_p2mp_route(tgen, "r1", "10.1.4.0/24", 24, "10.1.0.3", "r1-eth0")

    step("Remove neighbor filter configuration and verify")
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nno ip ospf neighbor-filter")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf neighbor-filter'", warn=False
    )
    assertmsg = "'ip ospf neighbor' not applied, but present in R1 configuration"
    assert rc, assertmsg

    step("Verify interface neighbor-filter is removed and neighbors present")
    verify_p2mp_interface(tgen, "r1", 3, 3, delay_reflood, "N/A")

    step("Add neighbor filter configuration and verify neighbors are filtered")
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nip ospf neighbor-filter nbr-filter")
    verify_p2mp_interface(tgen, "r1", 1, 1, delay_reflood, "nbr-filter")
    verify_p2mp_neighbor_missing(tgen, "r1", "2.2.2.2")
    verify_p2mp_neighbor_missing(tgen, "r1", "4.4.4.4")

    step("Remove nbr-filter prefix-list configuration to block r2 and verify neighbor")
    r1.vtysh_cmd("conf t\nno ip prefix-list nbr-filter seq 20")
    verify_p2mp_interface(tgen, "r1", 2, 2, delay_reflood, "nbr-filter")
    verify_p2mp_neighbor(
        tgen, "r1", "2.2.2.2", "Full/DROther", "10.1.0.2", "r1-eth0:10.1.0.1"
    )

    step("Delete nbr-filter prefix-list and verify neighbors are present")
    r1.vtysh_cmd("conf t\nno ip prefix-list nbr-filter")
    verify_p2mp_interface(tgen, "r1", 3, 3, delay_reflood, "N/A")


def test_p2mp_broadcast_neighbor_filter():
    p2mp_broadcast_neighbor_filter_common(False)


def test_p2mp_broadcast_neighbor_filter_delay_reflood():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Modify router r1 interface r1-eth0 p2mp delay-reflood configuration")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd(
        "conf t\ninterface r1-eth0\nip ospf network point-to-multipoint delay-reflood"
    )
    verify_p2mp_interface(tgen, "r1", 3, 3, True, "N/A")

    step("Modify router r2 interface r2-eth0 p2mp delay-reflood configuration")
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd(
        "conf t\ninterface r2-eth0\nip ospf network point-to-multipoint delay-reflood"
    )

    step("Modify router r3 interface r3-eth0 p2mp delay-reflood configuration")
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        "conf t\ninterface r3-eth0\nip ospf network point-to-multipoint delay-reflood"
    )

    step("Modify router r4 interface r4-eth0 p2mp delay-reflood configuration")
    r4 = tgen.gears["r4"]
    r4.vtysh_cmd(
        "conf t\ninterface r4-eth0\nip ospf network point-to-multipoint delay-reflood"
    )

    p2mp_broadcast_neighbor_filter_common(True)

    step("Recreate a partial P2MP mesh with neighbor filters")
    step("Add nbr-filter prefix-list configuration to block r4")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 30 permit any")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 10 deny 10.1.0.3/32")
    r1.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 20 deny 10.1.0.4/32")
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nip ospf neighbor-filter nbr-filter")

    r2.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 30 permit any")
    r2.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 10 deny 10.1.0.4/32")
    r2.vtysh_cmd("conf t\ninterface r2-eth0\nip ospf neighbor-filter nbr-filter")

    r3.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 30 permit any")
    r3.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 10 deny 10.1.0.1/32")
    r3.vtysh_cmd("conf t\ninterface r3-eth0\nip ospf neighbor-filter nbr-filter")

    r4.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 30 permit any")
    r4.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 10 deny 10.1.0.1/32")
    r4.vtysh_cmd("conf t\nip prefix-list nbr-filter seq 20 deny 10.1.0.2/32")
    r4.vtysh_cmd("conf t\ninterface r4-eth0\nip ospf neighbor-filter nbr-filter")

    step(
        "Add redistribution and spaced static routes to r1 to test delay flood retransmission"
    )
    r1.vtysh_cmd("conf t\nrouter ospf\nredistribute static")
    r1.vtysh_cmd("conf t\nip route 20.1.1.1/32 null0")
    sleep(1)
    r1.vtysh_cmd("conf t\nip route 20.1.1.2/32 null0")
    sleep(1)
    r1.vtysh_cmd("conf t\nip route 20.1.1.3/32 null0")
    sleep(1)
    r1.vtysh_cmd("conf t\nip route 20.1.1.4/32 null0")
    sleep(1)
    r1.vtysh_cmd("conf t\nip route 20.1.1.5/32 null0")
    sleep(1)

    step(
        "Verify the routes are installed on r1 with delay-reflood in P2MP partial mesh"
    )
    verify_p2mp_route(tgen, "r4", "20.1.1.1/32", 32, "10.1.0.3", "r4-eth0")
    verify_p2mp_route(tgen, "r4", "20.1.1.2/32", 32, "10.1.0.3", "r4-eth0")
    verify_p2mp_route(tgen, "r4", "20.1.1.3/32", 32, "10.1.0.3", "r4-eth0")
    verify_p2mp_route(tgen, "r4", "20.1.1.4/32", 32, "10.1.0.3", "r4-eth0")


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: ISC

# Copyright 2023 6WIND S.A.
# Authored by Farid Mihoub <farid.mihoub@6wind.com>
#
"""
BMP main module:
    - dissect monitoring messages in the way to get updated/withdrawed prefixes
    - XXX: missing RFCs references
    - XXX: more bmp messages types to dissect
    - XXX: complete bgp message dissection
"""
import ipaddress
import json
import os
import struct
import sys

from datetime import datetime

from bgp.update import BGPUpdate
from bgp.update.rd import RouteDistinguisher


SEQ = 0
LOG_DIR = "/var/log/"
LOG_FILE = "/var/log/bmp.log"

IS_ADJ_RIB_OUT = 1 << 4
IS_AS_PATH = 1 << 5
IS_POST_POLICY = 1 << 6
IS_IPV6 = 1 << 7
IS_FILTERED = 1 << 7

if not os.path.exists(LOG_DIR):
    os.makedirs(LOG_DIR)


def bin2str_ipaddress(ip_bytes, is_ipv6=False):
    if is_ipv6:
        return str(ipaddress.IPv6Address(ip_bytes))
    return str(ipaddress.IPv4Address(ip_bytes[-4:]))


def log2file(logs, log_file):
    """
    XXX: extract the useful information and save it in a flat dictionnary
    """
    with open(log_file, "a") as f:
        f.write(json.dumps(logs) + "\n")


def timestamp_print(message, file=sys.stderr):
    """Helper function to timestamp_print messages with timestamps."""

    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"[{current_time}] {message}", file=file)


# ------------------------------------------------------------------------------
class BMPCodes:
    """
    XXX: complete the list, provide RFCs.
    """

    VERSION = 0x3

    BMP_MSG_TYPE_ROUTE_MONITORING = 0x00
    BMP_MSG_TYPE_STATISTICS_REPORT = 0x01
    BMP_MSG_TYPE_PEER_DOWN_NOTIFICATION = 0x02
    BMP_MSG_TYPE_PEER_UP_NOTIFICATION = 0x03
    BMP_MSG_TYPE_INITIATION = 0x04
    BMP_MSG_TYPE_TERMINATION = 0x05
    BMP_MSG_TYPE_ROUTE_MIRRORING = 0x06
    BMP_MSG_TYPE_ROUTE_POLICY = 0x64

    # initiation message types
    BMP_INIT_INFO_STRING = 0x00
    BMP_INIT_SYSTEM_DESCRIPTION = 0x01
    BMP_INIT_SYSTEM_NAME = 0x02
    BMP_INIT_VRF_TABLE_NAME = 0x03
    BMP_INIT_ADMIN_LABEL = 0x04

    # peer types
    BMP_PEER_GLOBAL_INSTANCE = 0x00
    BMP_PEER_RD_INSTANCE = 0x01
    BMP_PEER_LOCAL_INSTANCE = 0x02
    BMP_PEER_LOC_RIB_INSTANCE = 0x03

    # peer header flags
    BMP_PEER_FLAG_IPV6 = 0x80
    BMP_PEER_FLAG_POST_POLICY = 0x40
    BMP_PEER_FLAG_AS_PATH = 0x20
    BMP_PEER_FLAG_ADJ_RIB_OUT = 0x10

    # peer loc-rib flag
    BMP_PEER_FLAG_LOC_RIB = 0x80
    BMP_PEER_FLAG_LOC_RIB_RES = 0x7F

    # statistics type
    BMP_STAT_PREFIX_REJ = 0x00
    BMP_STAT_PREFIX_DUP = 0x01
    BMP_STAT_WITHDRAW_DUP = 0x02
    BMP_STAT_CLUSTER_LOOP = 0x03
    BMP_STAT_AS_LOOP = 0x04
    BMP_STAT_INV_ORIGINATOR = 0x05
    BMP_STAT_AS_CONFED_LOOP = 0x06
    BMP_STAT_ROUTES_ADJ_RIB_IN = 0x07
    BMP_STAT_ROUTES_LOC_RIB = 0x08
    BMP_STAT_ROUTES_PER_ADJ_RIB_IN = 0x09
    BMP_STAT_ROUTES_PER_LOC_RIB = 0x0A
    BMP_STAT_UPDATE_TREAT = 0x0B
    BMP_STAT_PREFIXES_TREAT = 0x0C
    BMP_STAT_DUPLICATE_UPDATE = 0x0D
    BMP_STAT_ROUTES_PRE_ADJ_RIB_OUT = 0x0E
    BMP_STAT_ROUTES_POST_ADJ_RIB_OUT = 0x0F
    BMP_STAT_ROUTES_PRE_PER_ADJ_RIB_OUT = 0x10
    BMP_STAT_ROUTES_POST_PER_ADJ_RIB_OUT = 0x11

    # peer down reason code
    BMP_PEER_DOWN_LOCAL_NOTIFY = 0x01
    BMP_PEER_DOWN_LOCAL_NO_NOTIFY = 0x02
    BMP_PEER_DOWN_REMOTE_NOTIFY = 0x03
    BMP_PEER_DOWN_REMOTE_NO_NOTIFY = 0x04
    BMP_PEER_DOWN_INFO_NO_LONGER = 0x05
    BMP_PEER_DOWN_SYSTEM_CLOSED = 0x06

    # termincation message types
    BMP_TERM_TYPE_STRING = 0x00
    BMP_TERM_TYPE_REASON = 0x01

    # termination reason code
    BMP_TERM_REASON_ADMIN_CLOSE = 0x00
    BMP_TERM_REASON_UNSPECIFIED = 0x01
    BMP_TERM_REASON_RESOURCES = 0x02
    BMP_TERM_REASON_REDUNDANT = 0x03
    BMP_TERM_REASON_PERM_CLOSE = 0x04

    # policy route tlv
    BMP_ROUTE_POLICY_TLV_VRF = 0x00
    BMP_ROUTE_POLICY_TLV_POLICY = 0x01
    BMP_ROUTE_POLICY_TLV_PRE_POLICY = 0x02
    BMP_ROUTE_POLICY_TLV_POST_POLICY = 0x03
    BMP_ROUTE_POLICY_TLV_STRING = 0x04


# ------------------------------------------------------------------------------
class BMPMsg:
    """
    XXX: should we move register_msg_type and look_msg_type
    to generic Type class.
    """

    TYPES = {}
    UNKNOWN_TYPE = None
    HDR_STR = "!BIB"
    MIN_LEN = struct.calcsize(HDR_STR)
    TYPES_STR = {
        BMPCodes.BMP_MSG_TYPE_INITIATION: "initiation",
        BMPCodes.BMP_MSG_TYPE_PEER_DOWN_NOTIFICATION: "peer down notification",
        BMPCodes.BMP_MSG_TYPE_PEER_UP_NOTIFICATION: "peer up notification",
        BMPCodes.BMP_MSG_TYPE_ROUTE_MONITORING: "route monitoring",
        BMPCodes.BMP_MSG_TYPE_STATISTICS_REPORT: "statistics report",
        BMPCodes.BMP_MSG_TYPE_TERMINATION: "termination",
        BMPCodes.BMP_MSG_TYPE_ROUTE_MIRRORING: "route mirroring",
        BMPCodes.BMP_MSG_TYPE_ROUTE_POLICY: "route policy",
    }

    @classmethod
    def register_msg_type(cls, msgtype):
        def _register_type(subcls):
            cls.TYPES[msgtype] = subcls
            return subcls

        return _register_type

    @classmethod
    def lookup_msg_type(cls, msgtype):
        return cls.TYPES.get(msgtype, cls.UNKNOWN_TYPE)

    @classmethod
    def dissect_header(cls, data):
        """
        0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |    Version    |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                        Message Length                         |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Message Type  |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        """
        if len(data) < cls.MIN_LEN:
            pass
        else:
            _version, _len, _type = struct.unpack(cls.HDR_STR, data[0 : cls.MIN_LEN])
            return _version, _len, _type

    @classmethod
    def dissect(cls, data, log_file=None):
        global SEQ
        version, msglen, msgtype = cls.dissect_header(data)

        msg_data = data[cls.MIN_LEN : msglen]
        data = data[msglen:]

        if version != BMPCodes.VERSION:
            timestamp_print(
                f"Expected BMP version {BMPCodes.VERSION} but got version {version}."
            )
            return data

        msg_cls = cls.lookup_msg_type(msgtype)
        if msg_cls == cls.UNKNOWN_TYPE:
            timestamp_print(f"Got unknown message type ")
            return data

        timestamp_print(f"Got message type: {msg_cls}")

        msg_cls.MSG_LEN = msglen - cls.MIN_LEN
        logs = msg_cls.dissect(msg_data)
        logs["seq"] = SEQ
        log2file(logs, log_file if log_file else LOG_FILE)
        SEQ += 1

        return data


# ------------------------------------------------------------------------------
class BMPPerPeerMessage:
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |   Peer Type   | Peer Flags    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                Peer Address (16 bytes)                        |
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                          Peer AS                              |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                         Peer BGP ID                           |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                      Timestamp (seconds)                      |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                    Timestamp (microseconds)                   |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    PEER_UNPACK_STR = "!BB8s16sI4sII"
    PEER_TYPE_STR = {
        BMPCodes.BMP_PEER_GLOBAL_INSTANCE: "global instance",
        BMPCodes.BMP_PEER_RD_INSTANCE: "route distinguisher instance",
        BMPCodes.BMP_PEER_LOCAL_INSTANCE: "local instance",
        BMPCodes.BMP_PEER_LOC_RIB_INSTANCE: "loc-rib instance",
    }

    @classmethod
    def dissect(cls, data):
        (
            peer_type,
            peer_flags,
            peer_distinguisher,
            peer_address,
            peer_asn,
            peer_bgp_id,
            timestamp_secs,
            timestamp_microsecs,
        ) = struct.unpack_from(cls.PEER_UNPACK_STR, data)

        msg = {"peer_type": cls.PEER_TYPE_STR[peer_type]}

        if peer_type == 0x03:
            msg["is_filtered"] = bool(peer_flags & IS_FILTERED)
            msg["policy"] = "loc-rib"
        else:
            # peer_flags = 0x0000 0000
            # ipv6, post-policy, as-path, adj-rib-out, reserverdx4
            is_adj_rib_out = bool(peer_flags & IS_ADJ_RIB_OUT)
            is_as_path = bool(peer_flags & IS_AS_PATH)
            is_post_policy = bool(peer_flags & IS_POST_POLICY)
            is_ipv6 = bool(peer_flags & IS_IPV6)
            msg["policy"] = "post-policy" if is_post_policy else "pre-policy"
            msg["ipv6"] = is_ipv6
            msg["peer_ip"] = bin2str_ipaddress(peer_address, is_ipv6)

        peer_bgp_id = bin2str_ipaddress(peer_bgp_id)
        timestamp = float(timestamp_secs) + timestamp_microsecs * (10**-6)

        data = data[struct.calcsize(cls.PEER_UNPACK_STR) :]
        msg.update(
            {
                "peer_distinguisher": str(RouteDistinguisher(peer_distinguisher)),
                "peer_asn": peer_asn,
                "peer_bgp_id": peer_bgp_id,
                "timestamp": str(datetime.fromtimestamp(timestamp)),
            }
        )

        return data, msg


# ------------------------------------------------------------------------------
@BMPMsg.register_msg_type(BMPCodes.BMP_MSG_TYPE_ROUTE_MONITORING)
class BMPRouteMonitoring(BMPPerPeerMessage):
    @classmethod
    def dissect(cls, data):
        data, peer_msg = super().dissect(data)
        data, update_msg = BGPUpdate.dissect(data)
        return {**peer_msg, **update_msg}


# ------------------------------------------------------------------------------
class BMPStatisticsReport:
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                        Stats Count                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |         Stat Type             |          Stat Len             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                        Stat Data                              |
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    pass


# ------------------------------------------------------------------------------
@BMPMsg.register_msg_type(BMPCodes.BMP_MSG_TYPE_PEER_DOWN_NOTIFICATION)
class BMPPeerDownNotification(BMPPerPeerMessage):
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |    Reason     |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |            Data (present if Reason = 1, 2 or 3)               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    @classmethod
    def dissect(cls, data):
        data, peer_msg = super().dissect(data)

        msg = {
            **peer_msg,
            **{
                "bmp_log_type": "peer down",
            },
        }

        # XXX: dissect the bgp open message

        return msg


# ------------------------------------------------------------------------------
@BMPMsg.register_msg_type(BMPCodes.BMP_MSG_TYPE_PEER_UP_NOTIFICATION)
class BMPPeerUpNotification(BMPPerPeerMessage):
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                 Local Address (16 bytes)                      |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |          Local Port           |           Remote Port         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                    Sent OPEN Message                         #|
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                  Received OPEN Message                        |
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    UNPACK_STR = "!16sHH"
    MIN_LEN = struct.calcsize(UNPACK_STR)
    MSG_LEN = None

    @classmethod
    def dissect(cls, data):
        data, peer_msg = super().dissect(data)

        (local_addr, local_port, remote_port) = struct.unpack_from(cls.UNPACK_STR, data)

        msg = {
            **peer_msg,
            **{
                "bmp_log_type": "peer up",
                "local_ip": bin2str_ipaddress(local_addr, peer_msg.get("ipv6")),
                "local_port": int(local_port),
                "remote_port": int(remote_port),
            },
        }

        # XXX: dissect the bgp open message

        return msg


# ------------------------------------------------------------------------------
@BMPMsg.register_msg_type(BMPCodes.BMP_MSG_TYPE_INITIATION)
class BMPInitiation:
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |      Information Type         |        Information Length     |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                   Information (variable)                      |
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    TLV_STR = "!HH"
    MIN_LEN = struct.calcsize(TLV_STR)
    FIELD_TO_STR = {
        BMPCodes.BMP_INIT_INFO_STRING: "information",
        BMPCodes.BMP_INIT_ADMIN_LABEL: "admin_label",
        BMPCodes.BMP_INIT_SYSTEM_DESCRIPTION: "system_description",
        BMPCodes.BMP_INIT_SYSTEM_NAME: "system_name",
        BMPCodes.BMP_INIT_VRF_TABLE_NAME: "vrf_table_name",
    }

    @classmethod
    def dissect(cls, data):
        msg = {}
        while len(data) > cls.MIN_LEN:
            _type, _len = struct.unpack_from(cls.TLV_STR, data[0 : cls.MIN_LEN])
            _value = data[cls.MIN_LEN : cls.MIN_LEN + _len].decode()

            msg[cls.FIELD_TO_STR[_type]] = _value
            data = data[cls.MIN_LEN + _len :]

        return msg


# ------------------------------------------------------------------------------
class BMPTermination:
    """
    0 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8 1 2 3 4 5 6 7 8
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |          Information Type     |       Information Length      |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                 Information (variable)                        |
    ~                                                               ~
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    pass


# ------------------------------------------------------------------------------
class BMPRouteMirroring:
    pass


# ------------------------------------------------------------------------------
class BMPRoutePolicy:
    pass



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_isis_rlfa_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2020 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_isis_rlfa_topo1.py:

 +---------+                     +---------+
 |         |                     |         |
 |   RT1   |                     |   RT2   |
 |         +---------------------+         |
 |         |                     |         |
 +---+-----+                     +------+--+
     |                                  |
     |                                  |
     |                                  |
 +---+-----+                     +------+--+
 |         |                     |         |
 |   RT3   |                     |   RT4   |
 |         |                     |         |
 |         |                     |         |
 +---+-----+                     +------+--+
     |                                  |
     |                                  |
     |                                  |
 +---+-----+                     +------+--+
 |         |                     |         |
 |   RT5   |                     |   RT6   |
 |         |                     |         |
 |         |                     |         |
 +---+-----+                     +------+--+
     |                                  |
     |                                  |
     |                                  |
 +---+-----+                     +------+--+
 |         |                     |         |
 |   RT7   |                     |   RT8   |
 |         +---------------------+         |
 |         |                     |         |
 +---------+                     +---------+
"""

import os
import sys
import pytest
import json
import tempfile
from functools import partial

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.isisd, pytest.mark.ldpd]

# Global multi-dimensional dictionary containing all expected outputs
outputs = {}


def build_topo(tgen):
    "Build function"

    #
    # Define FRR Routers
    #
    for router in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "rt7", "rt8"]:
        tgen.add_router(router)

    #
    # Define connections
    #
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt2")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt1")
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt3")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt1")
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt4")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt2")
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt5")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt3")
    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt6")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt4")
    switch = tgen.add_switch("s6")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt5")
    switch = tgen.add_switch("s7")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt8")
    switch.add_link(tgen.gears["rt8"], nodeif="eth-rt6")
    switch = tgen.add_switch("s8")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt8")
    switch.add_link(tgen.gears["rt8"], nodeif="eth-rt7")

    #
    # Populate multi-dimensional dictionary containing all expected outputs
    #
    files = [
        "show_ip_route.ref",
        "show_ipv6_route.ref",
        "show_yang_interface_isis_adjacencies.ref",
    ]
    for rname in ["rt1"]:
        outputs[rname] = {}
        for step in range(1, 10 + 1):
            outputs[rname][step] = {}
            for file in files:
                if step == 1:
                    # Get snapshots relative to the expected initial network convergence
                    filename = "{}/{}/step{}/{}".format(CWD, rname, step, file)
                    outputs[rname][step][file] = open(filename).read()
                else:
                    if file == "show_yang_interface_isis_adjacencies.ref":
                        continue

                    # Get diff relative to the previous step
                    filename = "{}/{}/step{}/{}.diff".format(CWD, rname, step, file)

                    # Create temporary files in order to apply the diff
                    f_in = tempfile.NamedTemporaryFile(mode="w")
                    f_in.write(outputs[rname][step - 1][file])
                    f_in.flush()
                    f_out = tempfile.NamedTemporaryFile(mode="r")
                    os.system(
                        "patch -s -o %s %s %s" % (f_out.name, f_in.name, filename)
                    )

                    # Store the updated snapshot and remove the temporary files
                    outputs[rname][step][file] = open(f_out.name).read()
                    f_in.close()
                    f_out.close()


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    # For all registered routers, load the zebra configuration file
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_LDP, os.path.join(CWD, "{}/ldpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def router_compare_json_output(rname, command, reference):
    "Compare router JSON output"

    logger.info('Comparing router "%s" "%s" output', rname, command)

    tgen = get_topogen()
    expected = json.loads(reference)

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(topotest.router_json_cmp, tgen.gears[rname], command, expected)
    _, diff = topotest.run_and_expect(test_func, None, count=120, wait=0.5)
    assertmsg = '"{}" JSON output mismatches the expected result'.format(rname)
    assert diff is None, assertmsg


#
# Step 1
#
# Test initial network convergence
#
def test_isis_adjacencies_step1():
    logger.info("Test (step 1): check IS-IS adjacencies")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show yang operational-data /frr-interface:lib isisd",
            outputs[rname][1]["show_yang_interface_isis_adjacencies.ref"],
        )


def test_rib_ipv4_step1():
    logger.info("Test (step 1): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][1]["show_ip_route.ref"]
        )


def test_rib_ipv6_step1():
    logger.info("Test (step 1): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][1]["show_ipv6_route.ref"]
        )


#
# Step 2
#
# Action(s):
# -Configure rt8 (rt1's PQ router) to not accept targeted hello messages
#
# Expected changes:
# -All rt1 backup routes should be uninstalled
#
def test_rib_ipv4_step2():
    logger.info("Test (step 2): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Configuring rt8 to not accept targeted hello messages")
    tgen.net["rt8"].cmd(
        'vtysh -c "conf t" -c "mpls ldp" -c "address-family ipv4" -c "no discovery targeted-hello accept"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][2]["show_ip_route.ref"]
        )


def test_rib_ipv6_step2():
    logger.info("Test (step 2): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][2]["show_ipv6_route.ref"]
        )


#
# Step 3
#
# Action(s):
# -Configure rt8 (rt1's PQ router) to accept targeted hello messages
#
# Expected changes:
# -All rt1 previously uninstalled backup routes should be reinstalled
#
def test_rib_ipv4_step3():
    logger.info("Test (step 3): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Configuring rt8 to accept targeted hello messages")
    tgen.net["rt8"].cmd(
        'vtysh -c "conf t" -c "mpls ldp" -c "address-family ipv4" -c "discovery targeted-hello accept"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][3]["show_ip_route.ref"]
        )


def test_rib_ipv6_step3():
    logger.info("Test (step 3): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][3]["show_ipv6_route.ref"]
        )


#
# Step 4
#
# Action(s):
# -Disable RLFA on rt1's eth-rt2 interface
#
# Expected changes:
# -All non-ECMP routes whose primary nexthop is eth-rt2 should lose their backup nexthops
#
def test_rib_ipv4_step4():
    logger.info("Test (step 4): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Disabling RLFA on rt1's eth-rt2 interface")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "no isis fast-reroute remote-lfa tunnel mpls-ldp"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][4]["show_ip_route.ref"]
        )


def test_rib_ipv6_step4():
    logger.info("Test (step 4): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][4]["show_ipv6_route.ref"]
        )


#
# Step 5
#
# Action(s):
# -Disable RLFA on rt1's eth-rt3 interface
#
# Expected changes:
# -All non-ECMP routes whose primary nexthop is eth-rt3 should lose their backup nexthops
#
def test_rib_ipv4_step5():
    logger.info("Test (step 5): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Disabling RLFA on rt1's eth-rt3 interface")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt3" -c "no isis fast-reroute remote-lfa tunnel mpls-ldp"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][5]["show_ip_route.ref"]
        )


def test_rib_ipv6_step5():
    logger.info("Test (step 5): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][5]["show_ipv6_route.ref"]
        )


#
# Step 6
#
# Action(s):
# -Re-enable RLFA on rt1's eth-rt2 and eth-rt3 interfaces
#
# Expected changes:
# -Revert changes from the previous two steps (reinstall all backup routes)
#
def test_rib_ipv4_step6():
    logger.info("Test (step 6): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Re-enabling RLFA on rt1's eth-rt2 and eth-rt3 interfaces")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "isis fast-reroute remote-lfa tunnel mpls-ldp"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt3" -c "isis fast-reroute remote-lfa tunnel mpls-ldp"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][6]["show_ip_route.ref"]
        )


def test_rib_ipv6_step6():
    logger.info("Test (step 6): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][6]["show_ipv6_route.ref"]
        )


#
# Step 7
#
# Action(s):
# -Configure a PQ node prefix-list filter
#
# Expected changes:
# -All backup routes should be uninstalled
#
def test_rib_ipv4_step7():
    logger.info("Test (step 7): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Configuring a PQ node prefix-list filter")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute remote-lfa prefix-list PLIST"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][7]["show_ip_route.ref"]
        )


def test_rib_ipv6_step7():
    logger.info("Test (step 7): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][7]["show_ipv6_route.ref"]
        )


#
# Step 8
#
# Action(s):
# -Configure a prefix-list allowing rt8 as a PQ node
#
# Expected changes:
# -All backup routes should be installed again
#
def test_rib_ipv4_step8():
    logger.info("Test (step 8): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Configuring a prefix-list allowing rt8 as a PQ node")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "ip prefix-list PLIST seq 5 permit 10.0.255.8/32"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][8]["show_ip_route.ref"]
        )


def test_rib_ipv6_step8():
    logger.info("Test (step 8): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][8]["show_ipv6_route.ref"]
        )


#
# Step 9
#
# Action(s):
# -Change the maximum metric up to the PQ node to 30 on the eth-rt2 interface
#
# Expected changes:
# -All non-ECMP routes whose primary nexthop is eth-rt2 should lose their backup nexthops
#
def test_rib_ipv4_step9():
    logger.info("Test (step 9): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info(
        "Changing the maximum metric up to the PQ node to 30 on the eth-rt2 interface"
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "isis fast-reroute remote-lfa maximum-metric 30"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][9]["show_ip_route.ref"]
        )


def test_rib_ipv6_step9():
    logger.info("Test (step 9): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][9]["show_ipv6_route.ref"]
        )


#
# Step 10
#
# Action(s):
# -Change the maximum metric up to the PQ node to 40 on the eth-rt2 interface
#
# Expected changes:
# -All non-ECMP routes whose primary nexthop is eth-rt2 should recover their backup nexthops
#
def test_rib_ipv4_step10():
    logger.info("Test (step 10): verify IPv4 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info(
        "Changing the maximum metric up to the PQ node to 40 on the eth-rt2 interface"
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "isis fast-reroute remote-lfa maximum-metric 40"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ip route isis json", outputs[rname][10]["show_ip_route.ref"]
        )


def test_rib_ipv6_step10():
    logger.info("Test (step 10): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][10]["show_ipv6_route.ref"],
        )


# Memory leak test template
def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_table_direct_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2025 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_bgp_table_direct_topo1.py: Test the FRR PIM MSDP peer.
"""

import os
import sys
import json
from functools import partial
import re
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest

# Required to instantiate the topology builder class.
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

from lib.pim import McastTesterHelper

pytestmark = [pytest.mark.bgpd, pytest.mark.pimd]

app_helper = McastTesterHelper()


def build_topo(tgen):
    """
    +----+     +----+
    | r1 | <-> | r2 |
    +----+     +----+
       |
       |       +----+
       --------| r3 |
               +----+
    """

    # Create 3 routers
    for routern in range(1, 4):
        tgen.add_router(f"r{routern}")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_frr_config(os.path.join(CWD, f"{router.name}/frr.conf"))

    tgen.gears["r1"].run("ip link add blue type vrf table 10")
    tgen.gears["r1"].run("ip link set blue up")
    tgen.gears["r1"].run("ip link set r1-eth1 master blue")

    # Initialize all routers.
    tgen.start_router()

    app_helper.init(tgen)


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    app_helper.cleanup()
    tgen.stop_topology()


def expect_bgp_route(router, iptype, route, missing=False):
    "Wait until route is present on RIB for protocol."
    if missing:
        logger.info("waiting route {} go missing in {}".format(route, router))
    else:
        logger.info("waiting route {} in {}".format(route, router))

    tgen = get_topogen()
    expected_output = {route: [{"protocol": "bgp"}]}
    wait_time = 130
    if missing:
        expected_output = {route: None}
        wait_time = 5

    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears[router],
        "show {} route json".format(iptype),
        expected_output
    )

    _, result = topotest.run_and_expect(test_func, None, count=130, wait=1)
    assertmsg = f'"{router}" convergence failure'
    assert result is None, assertmsg


def test_bgp_convergence():
    "Wait for BGP protocol convergence"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for protocols to converge")

    # Wait for R2
    expect_bgp_route("r2", "ip", "10.254.254.1/32")
    expect_bgp_route("r2", "ip", "10.254.254.2/32")
    expect_bgp_route("r2", "ip", "10.254.254.3/32")

    # Wait for R3
    expect_bgp_route("r3", "ip", "10.254.254.1/32")
    expect_bgp_route("r3", "ip", "10.254.254.2/32")
    expect_bgp_route("r3", "ip", "10.254.254.3/32")


def test_route_change_convergence():
    "Change routes in table 2000 to test zebra redistribution."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r1"].vtysh_cmd("""
        configure terminal
        no ip route 10.254.254.2/32 lo table 2000
        ip route 10.254.254.10/32 lo table 2000
    """)

    # Check R2
    expect_bgp_route("r2", "ip", "10.254.254.2/32", missing=True)
    expect_bgp_route("r2", "ip", "10.254.254.10/32")

    # Check R3
    expect_bgp_route("r3", "ip", "10.254.254.2/32", missing=True)
    expect_bgp_route("r3", "ip", "10.254.254.10/32")


def test_configuration_removal_convergence():
    "Remove table direct configuration and check if routes went missing."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r1"].vtysh_cmd("""
        configure terminal
        router bgp 65001
         address-family ipv4 unicast
          no redistribute table-direct 2000
         exit-address-family
        exit

        router bgp 65001 vrf blue
         address-family ipv4 unicast
          no redistribute table-direct 2000
         exit-address-family
        exit
    """)

    # Check R2
    expect_bgp_route("r2", "ip", "10.254.254.1/32", missing=True)
    expect_bgp_route("r2", "ip", "10.254.254.3/32", missing=True)
    expect_bgp_route("r2", "ip", "10.254.254.10/32", missing=True)

    # Check R3
    expect_bgp_route("r3", "ip", "10.254.254.1/32", missing=True)
    expect_bgp_route("r3", "ip", "10.254.254.3/32", missing=True)
    expect_bgp_route("r3", "ip", "10.254.254.10/32", missing=True)


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
